content_handler = dict()
content_handler['document'] = 'resource/x-bb-document'
content_handler['externallink'] = 'resource/x-bb-externallink'
content_handler['folder'] = 'resource/x-bb-folder'
content_handler['courselink'] = 'resource/x-bb-courselink'
content_handler['forumlink'] = 'resource/x-bb-forumlink' # has not implemented get content for this
content_handler['blti-link'] = 'resource/x-bb-blti-link' # has not implemented get content for this
content_handler['file'] = 'resource/x-bb-file'
content_handler['asmt-test-link'] = 'resource/x-bb-asmt-test-link' # has not implemented get content for this
content_handler['assignment'] = 'resource/x-bb-assignment'
content_handler['blankpage'] = 'resource/x-bb-blankpage'



from bbcli.utils.utils import format_date
from bbcli.utils.error_handler import create_exception_handler, delete_exception_handler, list_exception_handler, update_exception_handler
import click
from bbcli.entities.content_builder_entitites import FileOptions, GradingOptions, StandardOptions, WeblinkOptions
from bbcli.services import contents_service
import time
import click
import threading
import concurrent.futures

from bbcli.entities.Node import Node
from bbcli.utils.URL_builder import URL_builder
from bbcli.utils import content_utils
from bbcli.utils.content_handler import content_handler
from bbcli.views import contents_view

url_builder = URL_builder()

base_url = 'https://ntnu.blackboard.com/learn/api/public/v1/'


def standard_options(function):
    function = click.option('-h', '--hide-content', is_flag=True,
                            help='Hide contents for students')(function)
    function = click.option(
        '-r', '--reviewable', is_flag=True, help='Make content reviewable')(function)
    function = click.option('--start-date', type=str,
                            help='When to make content available. Format: DD/MM/YY HH:MM:SS')(function)
    function = click.option(
        '--end-date', type=str, help='When to make content unavailable. Format: DD/MM/YY HH:MM:SS')(function)
    return function


def grading_options(function):
    function = click.option('-d', '--due-date', type=str,
                            help='Set a sumbission deadline for assignment. Format: DD/MM/YY HH:MM:SS')(function)
    function = click.option('-a', '--max-attempts', type=int,
                            help='Set a maximum amount of attempts.')(function)
    function = click.option('-u', '--unlimited-attempts',
                            is_flag=True, help='Enable unlimited attempts.')(function)
    function = click.option('-s', '--score', required=True,
                            type=int, help='Set assignment score reward.')(function)
    return function


def file_options(function):
    function = click.option('-n', '--new-window',
                            'launch_in_new_window', is_flag=True)(function)
    return function


def web_link_options(function):
    function = click.option('-n', '--new-window',
                            'launch_in_new_window', is_flag=True)(function)
    return function

@click.command(name='list', help='List the contents. Folders are blue and files are white.')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID, of the course you want to list content from.')
@click.option('-f', '--folders', required=False, is_flag=True, help='Specify this if you want to only list folders.')
@click.option('-ct', '--content-type', required=False, type=click.Choice(content_handler.keys(), case_sensitive=False))
@click.pass_context
@list_exception_handler
def list_contents(ctx, course_id: str, content_type, folders: bool = False):
    ct = 'content' if content_type is None else content_type
    click.echo(f'Listing the {ct}s...')
    start = time.time()

    response = contents_service.list_contents(ctx.obj['SESSION'], course_id)
    data = response.json()['results']
    folder_ids = []
    node_ids = []

    threads = []
    with concurrent.futures.ThreadPoolExecutor() as executor:
        for node in data:
            root = Node(node)
            worklist = [root]
            folder_ids.append(node['id'])
            args = [ctx, course_id, worklist, folder_ids, node_ids, root, folders, content_type]
            t = executor.submit(content_utils.list_contents_thread, *args)
            threads.append(t)
                
    for t in threads:
        root_node = t.result()
        if root_node is not None:
            contents_view.list_tree(root_node, folder_ids, node_ids)
        else: return 

    end = time.time()

    print(f'Fetch time: {end - start} seconds')


@click.command(name='get', help='Get a spesific content from a course, using the content id.')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID, of the course you want to get content from.')
@click.option('-co', '--content','node_id', required=True, type=str, help='CONTENT ID, of the content you want to fetch.')
@click.pass_context
@list_exception_handler
def get_content(ctx, course_id: str, node_id: str):
    content_utils.check_content_handler(ctx, course_id, node_id)
    

@click.command(name='attachment', help='Adds an attachment to a content. Only supports contents of type document and assignment')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID, of the course where the content is.')
@click.option('-co', '--content', 'content_id', required=True, type=str, help='CONTENT ID, of the content you want to attach a file to.')
@click.argument('file_path', required=True, type=click.Path(exists=True))
@click.pass_context
@create_exception_handler
def upload_attachment(ctx, course_id: str, content_id: str, file_path: str):
    contents_service.upload_attachment(ctx.obj['SESSION'], course_id, content_id, file_path)


@click.command(name='document', help='Creates a document content, optionally with file attachments')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID, of the course you want to create content.')
@click.option('-f', '--folder', 'parent_id', required=True, type=str, help='FOLDER ID, of the folder you want to create content in.')
@click.argument('title', required=True, type=str)
@click.argument('attachments', required=False, nargs=-1, type=click.Path())
@standard_options
@click.pass_context
@create_exception_handler
def create_document(ctx, course_id: str, parent_id: str, title: str, hide_content: bool, reviewable: bool, start_date: str=None, end_date: str=None, attachments: tuple=None):
    standard_options = StandardOptions(hide_content=hide_content, reviewable=reviewable)
    set_dates(standard_options, start_date, end_date)

    response = contents_service.create_document(
        ctx.obj['SESSION'], course_id, parent_id, title, standard_options, attachments)
    click.echo(response)


@click.command(name='file', help='Creates a file content.')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID, of the course where the content exists')
@click.option('-f', '--folder', 'parent_id', required=True, type=str, help='FOLDER ID, of the folder you want to create content in.')
@click.argument('title', required=True, type=str)
@click.argument('file_path', required=True, type=click.Path(exists=True))
@file_options
@standard_options
@click.pass_context
@create_exception_handler
def create_file(ctx, course_id: str, parent_id: str, title: str, file_path: str, 
                        launch_in_new_window:bool, hide_content: bool, reviewable: bool,
                        start_date: str=None, end_date: str=None):
    file_options = FileOptions(launch_in_new_window)
    standard_options = StandardOptions(
        hide_content=hide_content, reviewable=reviewable)
    set_dates(standard_options, start_date, end_date)
    response = contents_service.create_file(
        ctx.obj['SESSION'], course_id, parent_id, title, file_path, file_options, standard_options)
    click.echo(response)


@click.command(name='web-link', help='Create a web link content.')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID, of the course where the content exists')
@click.option('-f', '--folder', 'parent_id', required=True, type=str, help='FOLDER ID, of the folder you want to create content in.')
@click.argument('title', required=True, type=str)
@click.argument('url', required=True, type=str)
@standard_options
@web_link_options
@click.pass_context
@create_exception_handler
def create_web_link(ctx, course_id: str, parent_id: str, title: str, url: str, 
                        launch_in_new_window:bool, hide_content: bool, reviewable: bool,
                        start_date: str=None, end_date: str=None):
    web_link_options = WeblinkOptions(launch_in_new_window)
    standard_options = StandardOptions(hide_content, reviewable)
    set_dates(standard_options, start_date, end_date)
    response = contents_service.create_externallink(
        ctx.obj['SESSION'], course_id, parent_id, title, url, web_link_options, standard_options)
    click.echo(response)


@click.command(name='folder', help='Create a folder either in top level or inside another content.')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID, of the course where the content exists')
@click.option('-f', '--folder', 'parent_id', required=False, type=str, help='FOLDER ID, of the folder you want to create folder in.')
@click.argument('title', required=True, type=str)
@click.option('--is-bb-page', is_flag=True, help='Make folder a blackboard page')
@standard_options
@click.pass_context
@create_exception_handler
def create_folder(ctx, course_id: str, parent_id: str, title: str,
                        hide_content: bool, reviewable: bool, is_bb_page: bool = False,
                        start_date: str=None, end_date: str=None):
    standard_options = StandardOptions(hide_content, reviewable)
    set_dates(standard_options, start_date, end_date)
    response = contents_service.create_folder(
        ctx.obj['SESSION'], course_id, parent_id, title, is_bb_page, standard_options)
    click.echo(response)
    
@click.command(name='course-link', help='Create a course link content which redirects user to the target content.')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID, of the course where the content exists')
@click.option('-f', '--folder', 'parent_id', required=True, type=str, help='FOLDER ID, of the folder you want to create content in.')
@click.argument('title', required=True, type=str)
@click.argument('target_id', required=True, type=str)
@standard_options
@click.pass_context
@create_exception_handler
def create_courselink(ctx, course_id: str, parent_id: str, title: str, target_id: str,
                        hide_content: bool, reviewable: bool, 
                        start_date: str=None, end_date: str=None):
    standard_options = StandardOptions(hide_content, reviewable)
    set_dates(standard_options, start_date, end_date)
    response = contents_service.create_courselink(
        ctx.obj['SESSION'], course_id, parent_id, title, target_id, standard_options)
    click.echo(response)
    
@click.command(name='assignment', help='Creates an assignment.')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID, of the course where the content exists')
@click.option('-f', '--folder', 'parent_id', required=True, type=str, help='FOLDER ID, of the folder you want to create content in.')
@click.argument('title', required=True, type=str)
@click.argument('attachments', required=False, nargs=-1, type=click.Path())
@standard_options
@grading_options
@click.pass_context
@create_exception_handler
def create_assignment_from_contents(ctx, course_id: str, parent_id: str, title: str,
                      hide_content: bool, reviewable: bool,
                      start_date: str, end_date: str,
                      due_date: str, max_attempts: int, unlimited_attempts: bool, score: int,
                      attachments: tuple):
    """
    Creates an assignment.
    """
    standard_options = StandardOptions(hide_content, reviewable)
    grading_options = GradingOptions(
        attempts_allowed=max_attempts, is_unlimited_attemps_allowed=unlimited_attempts, score_possible=score)

    set_dates(standard_options, start_date, end_date)
    grading_options.due = format_date(due_date)

    response = contents_service.create_assignment(
        ctx.obj['SESSION'], course_id, parent_id, title, standard_options, grading_options, attachments)
    click.echo(response)


# TODO: ADD RESPONSES
@click.command(name='delete', help='Deletes a content.')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID, of the course where the content exists')
@click.option('-co', '--content', 'content_id', required=True, type=str, help='CONTENT ID, of the content you want to delete.')
@click.option('--delete-grades', is_flag=True, help='Deletes grades if a grade column is assosciated with the content.')
@click.pass_context
@delete_exception_handler
def delete_content(ctx, course_id: str, content_id: str, delete_grades: bool):
    response = contents_service.delete_content(ctx.obj['SESSION'], course_id, content_id, delete_grades)
    click.echo(response)

# TODO: ADD RESPONSES
@click.command(name='update', help='Updates a given content.\nEditable content types: document, files, assignments, externallinks, courselinks')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID, of the course where the content exists.')
@click.option('-co', '--content', 'content_id', required=True, type=str, help='CONTENT ID, of the content you want to update.')
@click.pass_context
@update_exception_handler
def update_content(ctx, course_id: str, content_id: str):
    response = contents_service.update_content(ctx.obj['SESSION'], course_id, content_id)
    click.echo(response)


"""
HELPER FUNCTIONS
"""


def set_dates(standard_options: StandardOptions, start_date: str, end_date: str):
    if start_date:
        standard_options.date_interval.start_date = format_date(start_date)
    if end_date:
        standard_options.date_interval.end_date = format_date(end_date)



# """Top-level package for RP To-Do."""
# # rptodo/__init__.py

# __app_name__ = "rptodo"
# __version__ = "0.1.0"

# (
#     SUCCESS,
#     DIR_ERROR,
#     FILE_ERROR,
#     DB_READ_ERROR,
#     DB_WRITE_ERROR,
#     JSON_ERROR,
#     ID_ERROR,
# ) = range(7)

# ERRORS = {
#     DIR_ERROR: "config directory error",
#     FILE_ERROR: "config file error",
#     DB_READ_ERROR: "database read error",
#     DB_WRITE_ERROR: "database write error",
#     ID_ERROR: "to-do id error",
# }


from io import StringIO


class StringBuilder:
    _file_str = None

    def __init__(self):
        self._file_str = StringIO()

    def append(self, str):
        self._file_str.write(str)

    def __str__(self):
        return self._file_str.getvalue()



from datetime import timezone
import json
import click
import requests
import dateutil.parser
from bbcli.services.contents_service import upload_file
from bbcli.services.utils.attempt_builder import AttemptBuilder
from tabulate import tabulate

from bbcli.utils.URL_builder import URL_builder

url_builder = URL_builder()


def get_assignments(session: requests.Session, course_id):
    url = url_builder.base_v2().add_courses().add_id(
        course_id).add_gradebook().add_columns().create()
    response = session.get(url)
    response.raise_for_status()
    response = json.loads(response.text)
    results = response['results']
    print_assignments(results)

# TODO: This should be in view
def print_assignments(assignments):
    for i in range(len(assignments)):
        column_id = assignments[i]['id']
        name = assignments[i]['name']
        due = 'N/A'
        if 'grading' in assignments[i]:
            if 'due' in assignments[i]['grading']:
                due = assignments[i]['grading']['due']
                due_datetime = utc_to_local(dateutil.parser.parse(due))
                date = str(due_datetime.date())
                time = str(due_datetime.time())
                due = f'{date} {time}'

        click.echo('{:<12s}{:<40s} due {:<10s}'.format(column_id, name, due))


def utc_to_local(utc_dt):
    return utc_dt.replace(tzinfo=timezone.utc).astimezone(tz=None)


def get_column_attempts(session: requests.Session, course_id, column_id, print_submitted):
    url = url_builder.base_v2().add_courses().add_id(course_id).add_gradebook(
    ).add_columns().add_id(column_id).add_attempts().create()

    response = session.get(url)
    response.raise_for_status()
    response = json.loads(response.text)
    results = response['results']

    if print_submitted:
        print_submitted_attempts(results)
    else:
        print_all_attempts(results)


def print_submitted_attempts(attempts):
    table = {'id': [], 'user id': [], 'status': [], 'score': [], 'created': []}
    statuses = ['NeedsGrading', 'Completed']
    for attempt in attempts:
        for status in statuses:
            if (status == attempt['status']):
                append_to_table(attempt, table)
                continue

    click.echo(tabulate(table, headers='keys'))


def print_all_attempts(attempts):
    table = {'id': [], 'user id': [], 'status': [], 'score': [], 'created': []}
    for attempt in attempts:
        append_to_table(attempt, table)
    click.echo(tabulate(table, headers='keys'))


def append_to_table(attempt, table):
    table['id'].append(attempt['id'])
    table['user id'].append(attempt['userId'])
    table['status'].append(attempt['status'])
    table['score'].append(
        attempt['score']) if 'score' in attempt else table['score'].append('N/A')
    created = utc_to_local(dateutil.parser.parse(attempt['created']))
    table['created'].append(created)


def get_column_attempt(session: requests.Session, course_id, column_id, attempt_id):
    url = url_builder.base_v2().add_courses().add_id(course_id).add_gradebook(
    ).add_columns().add_id(column_id).add_attempts().add_id(attempt_id).create()

    response = session.get(url)
    attempt = json.loads(response.text)
    attempt = json.dumps(attempt, indent=2)
    click.echo(attempt)


def create_column_attempt(session: requests.Session, course_id, column_id, studentComments=None, studentSubmission=None, dst: str = None, status=None, draft: bool = False):
    url = url_builder.base_v2().add_courses().add_id(course_id).add_gradebook(
    ).add_columns().add_id(column_id).add_attempts().create()

    if draft or dst is not None:
        status = 'InProgress'

    attempt = AttemptBuilder(studentComments=studentComments,
                             studentSubmission=studentSubmission,
                             status=status)
    data = attempt.create_json()
    json_data = json.dumps(data, indent=2)
    response = session.post(url, data=json_data)
    response_json = json.loads(response.text)
    click.echo(response_json)

    if dst is not None and response.status_code == 201:
        attempt_id = response_json['id']
        attach_file(session, course_id, attempt_id, dst)
        if draft:
            return
        update_column_attempt(session, course_id, column_id,
                              attempt_id, status='NeedsGrading')


def update_column_attempt(session: requests.Session, course_id, column_id, attempt_id, status=None, score=None, text=None, notes=None, feedback=None, studentComments=None, studentSubmission=None, exempt=None, dst=None):

    url = url_builder.base_v2().add_courses().add_id(course_id).add_gradebook(
    ).add_columns().add_id(column_id).add_attempts().add_id(attempt_id).create()

    attempt = AttemptBuilder(status=status, score=score, text=text,
                             notes=notes, feedback=feedback, studentComments=studentComments, studentSubmission=studentSubmission, exempt=exempt)
    data = attempt.create_json()
    json_data = json.dumps(data, indent=2)

    response = session.patch(url, data=json_data)
    response.raise_for_status()
    response = json.loads(response.text)
    click.echo(response)

    if dst is not None:
        attach_file(session, course_id, attempt_id, dst)


def attach_file(session: requests.Session, course_id, attempt_id, dst: str):
    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_gradebook().add_attempts().add_id(attempt_id).add_files().create()

    uploaded_file = upload_file(session, dst)

    data = {'uploadId': uploaded_file['id']}
    data = json.dumps(data)
    response = session.post(url, data)

    response_json = json.loads(response.text)
    click.echo(response_json)



import click

def print_courses(courses):
    click.echo('\n{:<12} {:<5}\n'.format('Id', 'Course Name'))
    for course in courses:
        course_id = course['id']
        name = course['name']
        click.echo('{:<12} {:<5}'.format(course_id, name))
    click.echo('\n\n')

def print_course(course):

    primary_id = course['id']
    course_id = course['courseId']
    name = course['name']

    click.echo('\n{:<12} {:<12}'.format('Id:', primary_id))
    click.echo('{:<12} {:<12}'.format('Course Id:', course_id))
    click.echo('{:<12} {:<12}\n'.format('Name:', name))


class AttemptBuilder(object):
    def __init__(self, status=None, score=None, text=None, notes=None, feedback=None, studentComments=None, studentSubmission=None, exempt=None) -> None:
        self.status = status
        self.score = score
        self.text = text
        self.notes = notes
        self.feedback = feedback
        self.studentComments = studentComments
        self.studentSubmission = studentSubmission
        self.exempt = exempt

    def create_json(self):
        attempt = {}
        if self.status:
            attempt.update({
                'status': self.status
            })
        if self.score:
            attempt.update({
                'score': self.score
            })
        if self.text:
            attempt.update({
                'text': self.text
            })
        if self.notes:
            attempt.update({
                'notes': self.notes
            })
        if self.feedback:
            attempt.update({
                'feedback': self.feedback
            })
        if self.studentComments:
            attempt.update({
                'studentComments': self.studentComments
            })
        if self.studentSubmission:
            attempt.update({
                'studentSubmission': self.studentSubmission
            })
        if self.exempt:
            attempt.update({
                'exempt': self.exempt
            })
        return attempt



from datetime import datetime
import click
from bbcli.entities.content_builder_entitites import DateInterval
from bbcli.services import announcements_service
from bbcli.utils.error_handler import create_exception_handler, delete_exception_handler, list_exception_handler, update_exception_handler
from bbcli.utils.utils import format_date
from bbcli.views import announcement_view
import os

@click.command(name='list', help='This command lists your announcements.\nEither all announcements, all announcements from a spesific course, or one announcement.')
@click.option('-c', '--course', 'course_id', required=False, type=str, help='COURSE ID, list announcements from a spesific course')
@click.option('-a', '--announcement', 'announcement_id', required=False, type=str, help='ANNONUCEMENT ID, list a spesific announcement from a course.')
@click.pass_context
@list_exception_handler
def list_announcements(ctx,course_id=None, announcement_id=None):
    response = None

    if announcement_id:
        response = announcements_service.list_announcement(
            ctx.obj['SESSION'], course_id, announcement_id)
        announcement_view.print_course_announcements([response])
    elif course_id:
        response = announcements_service.list_course_announcements(
            ctx.obj['SESSION'], course_id)
        announcement_view.print_course_announcements(response)
    else:
        user_name = os.getenv('BB_USERNAME')
        response = announcements_service.list_announcements(
            ctx.obj['SESSION'], user_name)
        announcement_view.print_announcements(response)


@click.command(name='create', help='Creates an announcement. Add --help for all options available.')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID of the course you want to create an announcement in.')
@click.argument('title', required=True, type=str)
@click.option('--start-date', type=str, help='When to make announcement available. Format: DD/MM/YY HH:MM:SS')
@click.option('--end-date', type=str, help='When to make announcement unavailable. Format: DD/MM/YY HH:MM:SS')
@click.pass_context
@create_exception_handler
def create_announcement(ctx, course_id: str, title: str, start_date: str, end_date: str):
    date_interval = DateInterval()
    if start_date or end_date:
        if start_date:
            date_interval.start_date = format_date(start_date)
        if end_date:
            date_interval.end_date = format_date(end_date)

    response = announcements_service.create_announcement(
        ctx.obj['SESSION'], course_id, title, date_interval)
    announcement_view.print_announcement_created(response)


@click.command(name='delete', help='Deletes an announcement. Add --help for all options available')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID of the course you want to create an announcement in.')
@click.option('-a', '--announcement', 'announcement_id', required=True, type=str, help='ANNOUNCEMENT ID, of the announcement you want to delete.')
@click.pass_context
@delete_exception_handler
def delete_announcement(ctx, course_id: str, announcement_id: str):
    announcements_service.delete_announcement(ctx.obj['SESSION'], course_id, announcement_id)
    announcement_view.print_announcement_deleted()


@click.command(name='update', help='Updates an announcement. Add --help for all options available.')
@click.option('-c', '--course', 'course_id', required=True, type=str, help='COURSE ID of the course you want to create an announcement in.')
@click.option('-a', '--announcement', 'announcement_id', required=True, type=str, help='ANNOUNCEMENT ID, of the annonucement you want to update.')
@click.pass_context
@update_exception_handler
def update_announcement(ctx, course_id: str, announcement_id: str):
    response = announcements_service.update_announcement(ctx.obj['SESSION'], course_id, announcement_id)
    announcement_view.print_announcement_updated(response)



import click
from bbcli.commands.contents import grading_options, set_dates, standard_options
from bbcli.entities.content_builder_entitites import GradingOptions, StandardOptions
from bbcli.services import assignment_service, contents_service
from bbcli.utils.error_handler import create_exception_handler, list_exception_handler, update_exception_handler
from bbcli.utils.utils import format_date


def attempt_options(function):
    function = click.option(
        '--status', default='Completed', help='The status of this attempt.', show_default=True)(function)
    function = click.option(
        '--text', help='The text grade associated with this attempt.')(function)
    function = click.option(
        '--score', type=int, help='The score associated with this attempt.')(function)
    function = click.option(
        '--notes', help='The instructor notes associated with this attempt.')(function)
    function = click.option(
        '--feedback', help='The instructor feedback associated with this attempt.')(function)
    function = click.option('--exempt', is_flag=True,
                            help='Whether the score associated with this attempt is ignored when computing the user\'s grade for the associated grade column.')(function)
    return function


# TODO: This function is a copy of the same function in contents.py. Fix this.
@click.command(name='create', help='Create an assignment.')
@click.option('-c', '--course' ,'course_id', required=True, help='COURSE ID, of the course you want to create an assignment in.')
@click.option('-f', '--folder', 'parent_id', required=True, type=str, help='FOLDER ID, of the folder you want to place the assignment.')
@click.argument('title', required=True, type=str)
@click.argument('attachments', required=False, nargs=-1, type=click.Path())
@standard_options
@grading_options
@click.pass_context
@create_exception_handler
def create_assignment(ctx, course_id: str, parent_id: str, title: str,
                      hide_content: bool, reviewable: bool,
                      start_date: str, end_date: str,
                      due_date: str, max_attempts: int, unlimited_attempts: bool, score: int,
                      attachments: tuple):
    standard_options = StandardOptions(hide_content, reviewable)
    grading_options = GradingOptions(
        attempts_allowed=max_attempts, is_unlimited_attemps_allowed=unlimited_attempts, score_possible=score)

    set_dates(standard_options, start_date, end_date)
    grading_options.due = format_date(due_date)

    response = contents_service.create_assignment(
        ctx.obj['SESSION'], course_id, parent_id, title, standard_options, grading_options, attachments)
    click.echo(response)


@click.command(name='list', help='List all assignments from a course.')
@click.option('-c', '--course', 'course_id', required=True, help='COURSE ID, of the course you want assignments from.')
@click.pass_context
@list_exception_handler
def get_assignments(ctx, course_id):
    assignment_service.get_assignments(ctx.obj['SESSION'], course_id)


@click.command(name='list', help='List attempts for an assignment.')
@click.option('-c', '--course' ,'course_id', required=True, help='COURSE ID, of the course you want the assignment attempts from')
@click.option('-a', '--assignment', 'column_id', required=True, help='ASSIGNMENT ID, of the assignment you want attempts from')
@click.option('--submitted', is_flag=True, help='List only submitted attempts.')
@click.pass_context
@list_exception_handler
def get_attempts(ctx, course_id, column_id, submitted):
    assignment_service.get_column_attempts(
        ctx.obj['SESSION'], course_id, column_id, print_submitted=submitted)


# TODO: Retrieve the submission w/ attachments.
@click.command(name='get', help='Get a specific attempt for an assignment.')
@click.option('-c', '--course' ,'course_id', required=True, help='COURSE ID, of the course of you want to get attempt from')
@click.option('-a', '--assignment', 'column_id', required=True, help='ASSIGNMENT ID, of the assignment you want attempts from')
@click.option('-at', '--attempt', 'attempt_id', required=True, help='ATTEMPT ID, of the attempt you want to fetch.')
@click.pass_context
@list_exception_handler
def get_attempt(ctx, course_id, column_id, attempt_id):
    assignment_service.get_column_attempt(
        ctx.obj['SESSION'], course_id, column_id, attempt_id)


@click.command(name='submit', help='Submit assignment attempt.')
@click.option('-c', '--course' ,'course_id', required=True, help='COURSE ID, of the course to submit an assignment to.')
@click.option('-a', '--assignment', 'column_id', required=True, help='ASSIGNMENT ID, of the assignment you want to submit to.')
@click.option('--studentComments', help='The student comments associated with this attempt.')
@click.option('--studentSubmission', help='The student submission text associated with this attempt.')
@click.option('--file', help='Attach a file to an attempt for a Student Submission. Relative path of file.')
@click.option('--draft', is_flag=True)
@click.pass_context
@create_exception_handler
def submit_attempt(ctx, course_id, column_id, studentComments, studentSubmission, file, draft):
    assignment_service.create_column_attempt(
        ctx.obj['SESSION'], course_id, column_id, studentComments=studentComments, studentSubmission=studentSubmission, dst=file, status='needsGrading', draft=draft)


@click.command(name='submit-draft', help='Submit assignment draft.')
@click.option('-c', '--course' ,'course_id', required=True, help='COURSE ID, of the course where the assignment is.')
@click.option('-a', '--assignment', 'column_id', required=True, help='ASSIGNMENT ID, of the assignment you want to submit to.')
@click.option('-at', '--attempt', 'attempt_id', required=True, help='ATTEMPT ID, of the attempt you want to update.')
@click.pass_context
@update_exception_handler
def submit_draft(ctx, course_id, column_id, attempt_id):
    assignment_service.update_column_attempt(
        ctx.obj['SESSION'], course_id=course_id, column_id=column_id, attempt_id=attempt_id, status='needsGrading')


@click.command(name='update', help='Update assignment.')
@click.option('-c', '--course' ,'course_id', required=True, help='COURSE ID, of the course where the assignment is.')
@click.option('-a', '--assignment', 'column_id', required=True, help='ASSIGNMENT ID, of the assignment you want to submit to.')
@click.option('-at', '--attempt', 'attempt_id', required=True, help='ATTEMPT ID, of the attempt you want to update.')
@attempt_options
@click.option('--studentComments', help='The student comments associated with this attempt.')
@click.option('--studentSubmission', help='The student submission text associated with this attempt.')
@click.pass_context
@update_exception_handler
def update_attempt(ctx, course_id, column_id, attempt_id, status, comments, submission, file):
    assignment_service.update_column_attempt(
        session=ctx.obj['SESSION'], course_id=course_id, column_id=column_id, attempt_id=attempt_id, status=status, studentComments=comments, studentSubmission=submission, dst=file)


@click.command(name='grade', help='Grade an assignment.')
@click.option('-c', '--course' ,'course_id', required=True, help='COURSE ID, of the course where the assignment is.')
@click.option('-a', '--assignment', 'column_id', required=True, help='ASSIGNMENT ID, of the assignment you want.')
@click.option('-at', '--attempt', 'attempt_id', required=True, help='ATTEMPT ID, of the attempt you want to grade.')
@attempt_options
@click.pass_context
@update_exception_handler
def grade_assignment(ctx, course_id, column_id, attempt_id, status, score, text, notes, feedback, exempt):
    if status is None:
        status = 'Completed'

    assignment_service.update_column_attempt(session=ctx.obj['SESSION'], status=status, course_id=course_id, column_id=column_id,
                                             attempt_id=attempt_id, score=score, text=text, notes=notes, feedback=feedback, exempt=exempt)



import os
from datetime import datetime
import mmap
from typing import Dict, List
from requests import Session
import html2text
import click

from bbcli.services.authorization_service import login

def check_valid_key(obj, key) -> bool:
    # print("the keys are", obj.keys())
    if key not in obj.keys():
        click.echo(f'The key: \"{key}\" is not in the object')
        return False
    else:
        return True


def check_response(response) -> bool:
    invalid_statuscodes = [401, 403, 404]
    if response.status_code in invalid_statuscodes:
        click.echo('Status: ' + str(response.json()['status']))
        click.echo('Message: ' + response.json()['message'])
        return False
    else:
        return True


def check_valid_date(cookie) -> bool:
    tmp = cookie
    start = int(tmp.find('expires')) + len('expires') + 1
    end = int(tmp.find(','))
    timestmp = int(tmp[start: end])
    expires = datetime.fromtimestamp(timestmp)
    now = datetime.now()
    if expires >= now:
        return True
    else:
        return False


def set_cookies(session: Session, cookies: List):
    for cookie in cookies:
        session.cookies.set(cookie['name'], cookie['value'])


def set_headers(session: Session, headers: List):
    for header in headers:
        session.headers.update(header)


def html_to_text(html_data: str):
    to_text = html2text.HTML2Text()
    return to_text.handle(html_data)


def input_body():
    MARKER = '# Everything below is ignored. Leave blank if you want empty body.\n'
    body = click.edit('\n\n' + MARKER)
    if body is not None:
        body = body.split(MARKER, 1)[0].rstrip('\n')
    return body


def format_date(date: str):
    if date:
        try:
            return datetime.strptime(date, '%d/%m/%y %H:%M:%S')
        except ValueError:
            click.echo('Value format is not valid, please see --help for more info.')
            raise click.Abort()

def get_download_path(file_name):
    """Returns the default downloads path for linux or windows"""
    if os.name == 'nt':
        import winreg
        sub_key = r'SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders'
        downloads_guid = '{374DE290-123F-4565-9164-39C4925E467B}'
        with winreg.OpenKey(winreg.HKEY_CURRENT_USER, sub_key) as key:
            location = winreg.QueryValueEx(key, downloads_guid)[0]
        return f'{location}\{file_name}'
    else:
        return os.path.join(os.path.expanduser('~'), f'Downloads/{file_name}')


def authorization_handler(func):
	def inner_function(*args, **kwargs):
		is_authorized = True if os.getenv("BB_ROUTER") != None else False
		if is_authorized:
			func(*args, **kwargs)
		else:
			click.echo('You are not logged in, running login script:')
			login()
			click.echo('You can now communicate with Blackboard LMS')
	return inner_function

def handle_fish_shell_completion():
    append_text = '_BB_COMPLETE=fish_source bb > ~/.config/fish/completions/bb.fish'
    path = os.path.join(os.path.expanduser('~'), f'/.config/fish/completions/bb.fish')
    if os.path.exists(path):
        with open(path, 'rb') as f, \
            mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as s:
            if s.find(bytearray(append_text)) != -1:
                is_activated = True
                click.echo('Shell completion already activated!')
    
    if is_activated == False:
        with open(path, 'a') as f:
            f.write(f'\n{append_text}\n')
            click.echo('Shell completion activated! Restart shell to load the changes.')


from datetime import datetime
import json
from subprocess import call
from typing import Dict, Any
import requests
from bbcli.entities.content_builder_entitites import DateInterval
from bbcli.services.courses_service import list_courses
from bbcli.utils.utils import input_body, set_cookies
import click

from bbcli.utils.URL_builder import URL_builder

url_builder = URL_builder()


def list_announcements(session: requests.Session, user_name: str):
    courses = list_courses(session, user_name=user_name)
    announcements = []

    for course in courses:
        url = url_builder.base_v1().add_courses().add_id(
            course['id']).add_announcements().create()
        course_announcements = session.get(url)
        course_announcements = json.loads(course_announcements.text)

        # Adds the course name to each course announcement list to make it easier to display which course the announcement comes from
        if 'results' in course_announcements:
            announcements.append({
                'course_name': course['name'],
                'course_announcements': course_announcements['results']
            })

    return announcements


def list_course_announcements(session: requests.Session, course_id: str):
    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_announcements().create()
    course_announcements = session.get(url)
    course_announcements.raise_for_status()
    course_announcements = json.loads(course_announcements.text)['results']
    return course_announcements


def list_announcement(session: requests.Session, course_id: str, announcement_id: str):
    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_announcements().add_id(announcement_id).create()
    announcement = session.get(url)
    announcement.raise_for_status()
    announcement = json.loads(announcement.text)
    return announcement

# TODO: Test if the duration actually makes it unavailable/available when it should


def create_announcement(session: requests.Session, course_id: str, title: str, date_interval: DateInterval):
    if title == '':
        raise click.BadParameter('Argument TITLE cannot be empty!')
    
    body = input_body()

    data = {
        'title': title,
        'body': body
    }
    if date_interval:
        start_date_str = datetime.strftime(
            date_interval.start_date, '%Y-%m-%dT%H:%m:%S.%fZ') if date_interval.start_date else None
        end_date_str = datetime.strftime(
            date_interval.end_date, '%Y-%m-%dT%H:%m:%S.%fZ') if date_interval.end_date else None

        data.update({
            'availability': {
                'duration': {
                    'start': start_date_str,
                    'end': end_date_str
                }
            }
        })

    data = json.dumps(data)

    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_announcements().create()
    response = session.post(url, data=data)
    response.raise_for_status()

    return response.text


def delete_announcement(session: requests.Session, course_id: str, announcement_id: str):
    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_announcements().add_id(announcement_id).create()
    response = session.delete(url)
    response.raise_for_status()
    return response.text


def update_announcement(session: requests.Session, course_id: str, announcement_id: str):

    announcement = list_announcement(
        session=session, course_id=course_id, announcement_id=announcement_id)
    MARKER = '# Everything below is ignored.\n'
    editable_data = {
        'title': announcement['title'],
        'body': announcement['body'],
        'created': announcement['created'],
        'availability': announcement['availability'],
        'draft': announcement['draft']
    }
    announcement = json.dumps(editable_data, indent=2)
    new_data = click.edit(announcement + '\n\n' + MARKER)

    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_announcements().add_id(announcement_id).create()
    response = session.patch(url, data=new_data)
    response.raise_for_status()

    return response.text



from anytree import Node as Nd, RenderTree
from colorama import Fore, Style
from bbcli.utils.utils import html_to_text
import click
import tempfile, os
from subprocess import call

def list_tree(root, folder_ids, node_ids):
    # color = Fore.RESET if only_folders else Fore.BLUE
    color = Fore.BLUE
    for pre, fill, node in RenderTree(root):
        node_id = None
        if 'node_id' in dir(node):
            node_id = getattr(node, 'node_id')
        else:
            click.echo('Could not retreive the node id.')
        if node_id in folder_ids:
            click.echo(f'{pre}{color}{node_id} {node.name} {Style.RESET_ALL}')
        elif node_id in node_ids:
            click.echo(f'{pre}{node_id} {node.name}')
        else:
            click.echo('Neither node nor folder.')
        

def open_vim(data):
    str = data['title'] + '\n'
    str += html_to_text(data['body'])

    EDITOR = os.environ.get('EDITOR','vim') #that easy!

    initial_message = bytearray(str, encoding='utf8')

    with tempfile.NamedTemporaryFile(suffix=".tmp") as tf:
        tf.write(initial_message)
        tf.flush()
        call([EDITOR, tf.name])

        # do the parsing with `tf` using regular File operations.
        # for instance:
        # tf.seek(0)
        # edited_message = tf.read()
        # print (edited_message.decode("utf-8"))

def open_less_page(str):
    import pydoc
    pydoc.pager(str)





from http.client import INTERNAL_SERVER_ERROR
from typing import Dict, List
import requests
import click
from http import HTTPStatus

id_parameter_names = ['course_id', 'announcement_id','content_id', 'node_id', 'parent_id', 'target_id', 'column_id']

def get_used_parameters(kwargs: Dict, only_id_parameters: bool=False, no_id_parameters:bool= False):
    params = kwargs.keys()
    key_list = []
    for param in params:
        if only_id_parameters:
            id_used(kwargs, param, key_list)
        elif no_id_parameters:
            param_used(kwargs, param, key_list)
        else:
            param_and_id_used(kwargs, param, key_list)
    return key_list

def id_used(kwargs: Dict, param: str, key_list: List):
    if kwargs[param] != None and param in id_parameter_names:
        key_list.append(param)

def param_used(kwargs: Dict, param: str, key_list: List):
    if kwargs[param] != None and param not in id_parameter_names:
        key_list.append(param)

def param_and_id_used(kwargs: Dict, param: str, key_list: List):
    if kwargs[param] != None:
        key_list.append(param)

# Handles exceptions where a resource is not found or user don't have the right permission
def list_exception_handler(func):
    def inner_function(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except requests.exceptions.HTTPError as err:
            click.echo(err.response.text + '\n\n')
            status_code = err.response.json()['status']
            forbidden_handler(status_code)
            not_found_handler(status_code, kwargs)
            bad_request_or_not_found_handler(status_code, kwargs)
            internal_server_error_handler(status_code)
    return inner_function

def create_exception_handler(func):
    def inner_function(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except requests.exceptions.HTTPError as err:
            status_code = err.response.json()['status']
            click.echo(err.response.text + '\n\n')
            forbidden_handler(status_code)
            not_found_handler(status_code, kwargs)
            bad_request_or_not_found_handler(status_code, kwargs)
            insufficient_storage_handler(status_code)
            conflict_handler(status_code)
            internal_server_error_handler(status_code)
            unprocessable_entity(status_code)
    return inner_function

def delete_exception_handler(func):
    def inner_function(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except requests.exceptions.HTTPError as err:
            status_code = err.response.json()['status']
            click.echo(err.response.text + '\n\n')
            forbidden_handler(status_code)
            bad_request_or_not_found_handler(status_code, kwargs)
            internal_server_error_handler(status_code)
    return inner_function

def update_exception_handler(func):
    def inner_function(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except requests.exceptions.HTTPError as err:
            status_code = err.response.json()['status']
            click.echo(err.response.text + '\n\n')
            forbidden_handler(status_code)
            not_found_handler(status_code, kwargs)
            bad_request_or_not_found_handler(status_code, kwargs)
            internal_server_error_handler(status_code)
            insufficient_storage_handler(status_code)
    return inner_function


def forbidden_handler(status_code):
    if status_code == HTTPStatus.FORBIDDEN:
        click.echo('Current user do not have the right permissions to use/edit this resource/resources.')
        raise click.Abort()

def bad_request_or_not_found_handler(status_code, kwargs):
    if status_code == HTTPStatus.BAD_REQUEST or status_code == HTTPStatus.NOT_FOUND:
        click.echo(f'Bad parameters! Please check if the id/-s of type: {get_used_parameters(kwargs, only_id_parameters=True)} is correct')
        params_only_list = get_used_parameters(kwargs, no_id_parameters=True)
        if len(params_only_list) != 0:
            click.echo(f'and if all all of the parameters of type: {get_used_parameters(kwargs, no_id_parameters=True)} have correct format.')
        click.echo('HINT: use --help for more info')
        raise click.Abort()

def not_found_handler(status_code, kwargs):
    if status_code == HTTPStatus.NOT_FOUND:
        click.echo(f'Could not find resource/-s with the id/-s of type: {get_used_parameters(kwargs, only_id_parameters=True)}.')
        raise click.Abort()
    
def insufficient_storage_handler(status_code):
    if status_code == HTTPStatus.INSUFFICIENT_STORAGE:
        click.echo(f'This folder cannot hold any more children. Folder quota exceeded.')
        raise click.Abort()

def conflict_handler(status_code):
    if status_code == HTTPStatus.CONFLICT:
        click.echo('A conflict have appeared!')
        raise click.Abort()

def internal_server_error_handler(status_code):
    if status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        click.echo('An internal server error has appered! HINT: Check --help for correct formatting of parameters.')
        click.echo('Remember id have the format _x_1, where you replace the x with a number.')
        raise click.Abort()

def unprocessable_entity(status_code):
    if status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        click.echo('This file is potentially unsafe as determined from an XSS scan.')
        raise click.Abort()


import click
from typing import List
from bbcli.utils.utils import html_to_text


def print_announcements(announcements: List):
    announcements.reverse()
    for course in announcements:
        print_course_announcements(course['course_announcements'], course['course_name'])

def print_course_announcements(course_announcements: List, course_name: str = None):
    course_announcements.reverse()
    for announcement in course_announcements:
        if 'body' in announcement:
            announcement_id = announcement['id']
            title = announcement['title']
            body = html_to_text(announcement['body'])
            created = announcement['created'].split('T')[0]

            click.echo('----------------------------------------------------------------------\n')
            if course_name:
                click.echo(f'{course_name}\n')
            click.echo('{:<15} {:<15}'.format('Id: ', announcement_id))
            click.echo('{:<15} {:<15}'.format('Title: ', title))
            click.echo('{:<15} {:<15}'.format('Date: ', created))
            click.echo('\n{:<15}\n'.format(body))

def print_announcement_created(announcement):
    click.echo('\nAnnouncement sucessfully created:\n\n' + announcement)


def print_announcement_deleted():
    click.echo('\nAnnouncement sucessfully deleted.\n')

def print_announcement_updated(announcement):
    click.echo('\nAnnouncement sucessfully updated:\n\n' + announcement)



from typer.testing import CliRunner

from bbcli import __app_name__, __version__, cli, get_user

runner = CliRunner()

def test_version():
	# invokes the command line and checks the version
	result = runner.invoke(cli.app, ['--version'])
	assert result.exit_code == 0

	assert result.output == f"{__app_name__} v{__version__}\n"
	assert f"{__app_name__} v{__version__}\n" in result.stdout


def test_endpoint_getuser():
	result = runner.invoke(get_user, input='hanswf')
	assert not result.exception 




from urllib import request
import os
from bbcli.entities.RequestData import RequestData
import requests
import json
from bs4 import BeautifulSoup
import time
from getpass import getpass

ctx = ''
session_id = ''
canary = ''
auth_method_id = ''
hpg_request_id = ''
flowtoken = ''
login_username = ''
otc = None


def login():
    print("Logging in...")
# TODO: Let user choose between feide log in or ID-gate
    feide_login()

# -------------------- FEIDE LOGIN PART -------------------- #


def feide_login():
    session = requests.Session()
    request_data = RequestData()

    scrape_login_page(session, request_data)
    response = post_login_request(session, request_data, get_credentials())
    scrape_login_response(response, request_data)
    response = post_adfs_request(session, request_data)
    scrape_adfs_response(response, request_data)
    response = post_microsoft_login(session, request_data)
    scrape_microsoft_login_response(response, request_data)

    begin_auth(session, request_data)
    response = process_auth(session, request_data, otc)
    scrape_process_auth(response, request_data)
    post_auth_saml_SSO(session, request_data)

    write_to_env_data(session)
    print('Login successful!')


def get_credentials():
    # TODO: Give user an option to save their username and password
    # so they don't have to write their login everytime they want to login
    username = getpass("Username: ")
    password = getpass()

    credentials = {
        'username': username,
        'password': password
    }

    return credentials


def scrape_login_page(session, request_data):
    html = session.get('http://innsida.ntnu.no/blackboard')
    soup = BeautifulSoup(html.text, 'lxml')
    auth_state = soup.find('input', {'name': 'AuthState'})['value']

    request_data.params = (
        ('org', 'ntnu.no'),
        ('AuthState', auth_state)
    )


def post_login_request(session, request_data, credentials):
    request_data.data = {
        'feidename': credentials['username'],
        'password': credentials['password']
    }
    global login_username
    login_username = credentials['username']

    return session.post('https://idp.feide.no/simplesaml/module.php/feide/login', params=request_data.params, data=request_data.data)


def scrape_login_response(response, request_data):
    soup = BeautifulSoup(response.text, 'lxml')
    try:
        samlResponse = soup.find(
            'input', {'name': 'SAMLResponse'})['value']
        relayState = soup.find(
            'input', {'name': 'RelayState'})['value']
    except TypeError:
        raise TypeError("Username or password is wrong...")
    else:
        request_data.data = {
            'SAMLResponse': samlResponse,
            'RelayState': relayState
        }


def post_adfs_request(session, request_data):
    return session.post('https://adfs.ntnu.no/adfs/ls/', data=request_data.data)


def scrape_adfs_response(response, request_data):
    soup = BeautifulSoup(response.text, 'lxml')

    wa = soup.find('input', {'name': 'wa'})['value']

    wresult = soup.find('input', {'name': 'wresult'})['value']
    wctx = soup.find('input', {'name': 'wctx'})['value']

    request_data.data = {
        'wa': wa,
        'wresult': wresult,
        'wctx': wctx
    }


def post_microsoft_login(session, request_data):
    return session.post(
        'https://login.microsoftonline.com/login.srf', data=request_data.data)


def scrape_microsoft_login_response(response, request_data):
    soup = BeautifulSoup(response.text, 'lxml')

    script_tag = soup.find('script')
    # Remove certaint parts of the script content to be able to use json.loads
    script_tag = script_tag.contents[0][20:len(script_tag.contents[0]) - 7]
    script_content = json.loads(script_tag)

    global ctx
    global session_id
    global canary
    global auth_method_id
    global hpg_request_id

    ctx = script_content['sCtx']
    flowtoken = script_content['sFT']
    session_id = script_content['sessionId']
    canary = script_content['canary']
    auth_methods = script_content['arrUserProofs']
    auth_method_id = find_default_auth_method(auth_methods)
    if(auth_method_id == None):
        auth_method_id = choose_auth_method(auth_methods)
    hpg_request_id = response.headers['x-ms-request-id']

    request_data.data = {
        "AuthMethodId": auth_method_id,
        "Method": "BeginAuth",
        "ctx": ctx,
        "flowToken": flowtoken
    }
    request_data.data = json.dumps(request_data.data)


def find_default_auth_method(auth_methods):
    for auth_method in auth_methods:
        if auth_method['isDefault'] == True:
            return auth_method['authMethodId']


def choose_auth_method(auth_methods):
    print('Choose auth method:\n')
    choice = 1
    for auth_method in auth_methods:
        print(str(choice) + ". " + auth_method['authMethodId'])
        choice += 1

    user_input = input()
    return auth_methods[int(user_input) - 1]['authMethodId']


def begin_auth(session, request_data):
    global type

    if auth_method_id == 'PhoneAppNotification':
        begin_phone_app_auth(session, request_data)
    elif auth_method_id == 'OneWaySMS' or auth_method_id == 'PhoneAppOTP':
        begin_one_time_code_auth(session, request_data)


def begin_phone_app_auth(session, request_data):
    response = session.post(
        'https://login.microsoftonline.com/common/SAS/BeginAuth', data=request_data.data)

    j = json.loads(response.text)
    app_auth_wait(session, request_data, j['CorrelationId'], j['FlowToken'])


def begin_one_time_code_auth(session, request_data):
    global otc
    global flowtoken

    response = session.post(
        'https://login.microsoftonline.com/common/SAS/BeginAuth', data=request_data.data)

    SMS_code = input('Enter code: ')
    otc = SMS_code
    j = json.loads(response.text)

    request_data.data = {
        'AdditionalAuthData': SMS_code,
        'AuthMethodId': auth_method_id,
        'Ctx': j['Ctx'],
        'FlowToken': j['FlowToken'],
        'Method': 'EndAuth',
        'PollCount': 1,
        'SessionId': session_id
    }
    request_data.headers = {
        'client-request-id': j['CorrelationId']
    }

    data = json.dumps(request_data.data)

    response = session.post(
        'https://login.microsoftonline.com/common/SAS/EndAuth', data=data, headers=request_data.headers)
    j = json.loads(response.text)
    flowtoken = j['FlowToken']


def app_auth_wait(session, request_data, client_request_id, flow_token):
    global flowtoken

    request_data.data = {
        "Method": "EndAuth",
        "SessionId": session_id,
        "FlowToken": flow_token,
        "Ctx": ctx,
        "AuthMethodId": auth_method_id,
        "PollCount": 1
    }
    request_data.headers = {
        'client-request-id': client_request_id,
    }

    data = json.dumps(request_data.data)

    for tries in range(10):
        response = session.post(
            'https://login.microsoftonline.com/common/SAS/EndAuth', data=data, headers=request_data.headers)
        j = json.loads(response.text)
        if j['Success'] == True:
            print('User accepted')
            flowtoken = j['FlowToken']
            break
        request_data.data = {
            "Method": "EndAuth",
            "SessionId": session_id,
            "FlowToken": j['FlowToken'],
            "Ctx": ctx,
            "AuthMethodId": auth_method_id,
            "PollCount": str(tries + 1)
        }
        data = json.dumps(request_data.data)
        print('User still havent authorized')
        time.sleep(2)


def process_auth(session, request_data, otc=None):

    request_data.data = {
        'request': ctx,
        'mfaAuthMethod': auth_method_id,
        'canary': canary,
        'login': login_username,
        'flowToken': flowtoken,
        'hpgrequestid': hpg_request_id,
        'otc': otc
    }

    return session.post('https://login.microsoftonline.com/common/SAS/ProcessAuth', data=request_data.data)


def scrape_process_auth(response, request_data):
    soup = BeautifulSoup(response.text, 'lxml')

    saml_response = soup.find('input', {'name': 'SAMLResponse'})['value']

    request_data.data = {
        'SAMLResponse': saml_response
    }


def post_auth_saml_SSO(session, requets_data):
    return session.post(
        'https://ntnu.blackboard.com/auth-saml/saml/SSO', data=requets_data.data)


def write_to_env_data(session):
    BB_ROUTER = session.cookies['BbRouter']
    bb_router_values = BB_ROUTER.split(',')
    xsrf = bb_router_values[len(bb_router_values) - 1].split(':')
    xsrf_value = xsrf[1]

    enviroment_path = f'{os.path.dirname(os.path.abspath(__file__))}/../'

    f = open(f'{enviroment_path}.env', 'w')
    f.write(f'BB_ROUTER={BB_ROUTER}\n')
    f.write(f'XSRF={xsrf_value}\n')
    f.write(f'BB_USERNAME={login_username}\n')
    f.close()


if __name__ == "__main__":
    login()




import requests
from bbcli.utils.utils import handle_fish_shell_completion, set_cookies, set_headers
from bbcli import __app_name__, __version__
import os
import shutil
from dotenv import load_dotenv
from bbcli import check_valid_date
import click

from bbcli.commands.courses import list_courses
from bbcli.commands.announcements import list_announcements, create_announcement, delete_announcement, update_announcement
from bbcli.commands.contents import create_assignment_from_contents, create_courselink, create_folder, delete_content, list_contents, create_document, create_file, create_web_link, update_content, upload_attachment, get_content
from bbcli.commands.assignments import get_assignments, submit_attempt, grade_assignment, get_attempts, get_attempt, submit_draft, update_attempt, submit_draft, create_assignment
from bbcli.services.authorization_service import login
import mmap


def initiate_session():
    bb_cookie = {
        'name': 'BbRouter',
        'value': os.getenv("BB_ROUTER")
    }
    xsrf = {'X-Blackboard-XSRF': os.getenv('XSRF')}

    session = requests.Session()
    set_cookies(session, [bb_cookie])
    set_headers(session, [xsrf])
    session.headers.update({'Content-Type': 'application/json'})
    return session

def authenticate_user():
    load_dotenv()
    bb_cookie = os.getenv('BB_ROUTER')
    is_authorized = True if bb_cookie != None and check_valid_date(bb_cookie) else False
    if not is_authorized:
        click.echo('You are not logged in. Executing authorization script...')
        login()
    

@click.group()
@click.pass_context
@click.version_option(__version__)
def entry_point(ctx):
    ctx.ensure_object(dict)

"""
LOGIN AND LOGOUT COMMANDS
"""
@click.command(name='login')
def authorize_user():
    """
    Authorize user with username and password.
    """
    login()

@click.command(name='logout')
def logout():
    """
    Logout user.
    """
    open(f'{os.path.dirname(os.path.abspath(__file__))}/.env', 'w').close()
    click.echo('Sucessfully logged out.')

entry_point.add_command(authorize_user)
entry_point.add_command(logout)

"""
SHELL COMPLETION COMMANDS
"""
@click.command(name='activate-shell-completion', help='Activate shell completion for your shell')
@click.argument('shell', required=True, type=str)
def activate_shell_completion(shell: str):
    if shell == 'fish':
        handle_fish_shell_completion()
    else:
        if shell == 'bash' or shell == 'zsh':
            is_activated = False
            path = os.path.join(os.path.expanduser('~'), f'.{shell}rc')
            append_text = f'. ~/.bb-complete.{shell}'
            with open(path, 'rb') as f, \
                mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as s:
                if s.find(bytearray(append_text.encode())) != -1:
                    is_activated = True

            if not is_activated:
                shutil.copy(f'{os.path.dirname(os.path.abspath(__file__))}/shell-completion/.bb-complete.bash', os.path.expanduser('~'))
                with open(os.path.join(os.path.expanduser('~'), f'.{shell}rc'), 'a') as f:
                    f.write('\n. ~/.bb-complete.bash\n')
                    click.echo('Shell completion activated! Restart shell to load the changes.')
            else:
                click.echo('Shell completion already activated.')
        else: 
            click.echo('Did not recognize the Shell or CLI shell completion is not compatible with your Shell.')


entry_point.add_command(activate_shell_completion)

"""
COURSE COMMANDS ENTRY POINT
"""

@entry_point.group()
@click.pass_context
def courses(ctx):
    """
    Commands for listing courses
    """
    authenticate_user()
    load_dotenv()
    session = initiate_session()
    ctx.obj['SESSION'] = session


courses.add_command(list_courses)


"""
ANNOUNCEMENT COMMANDS ENTRY POINT
"""


@entry_point.group()
@click.pass_context
def announcements(ctx):
    """
    Commands for listing, creating, deleting and updating announcements
    """
    authenticate_user()
    load_dotenv()
    session = initiate_session()
    ctx.obj['SESSION'] = session

announcements.add_command(list_announcements)
announcements.add_command(create_announcement)
announcements.add_command(delete_announcement)
announcements.add_command(update_announcement)


@entry_point.group()
@click.pass_context
def assignments(ctx):
    """
    Commands for creating, listing and submitting assignments.
    """
    authenticate_user()
    load_dotenv()
    session = initiate_session()
    ctx.obj['SESSION'] = session

assignments.add_command(get_assignments)
assignments.add_command(create_assignment)
assignments.add_command(grade_assignment)


@assignments.group()
@click.pass_context
def attempts(ctx):
    """
    Commands for creating, submitting and listing attempts for an assignment.
    """
    pass


attempts.add_command(get_attempts)
attempts.add_command(get_attempt)
attempts.add_command(submit_attempt)
attempts.add_command(submit_draft)
attempts.add_command(update_attempt)

"""
CONTENT COMMANDS ENTRY POINT
"""


@entry_point.group()
@click.pass_context
def contents(ctx):
    """
    Commands for listing, creating, deleting, updating and downloading content
    """
    authenticate_user()
    load_dotenv()
    session = initiate_session()
    ctx.obj['SESSION'] = session


contents.add_command(list_contents)
contents.add_command(get_content)
contents.add_command(delete_content)
contents.add_command(update_content)

"""
CONTENTS CREATE COMMANDS ENTRY POINT
"""


@contents.group()
@click.pass_context
def create(ctx):
    """
    Commands for creating different types of content types in blackboard
    """
    authenticate_user()
    load_dotenv()
    session = initiate_session()
    ctx.obj['SESSION'] = session


create.add_command(create_document)
create.add_command(create_file)
# create.add_command(create_web_link)
create.add_command(create_folder)
create.add_command(create_courselink)
create.add_command(upload_attachment)
create.add_command(create_assignment_from_contents)



class RequestData:
    def __init__(self, data=None, params=None, headers=None, cookies=None):
        self.data = data
        self.params = params
        self.headers = headers
        self.cookies = cookies



from collections import defaultdict, deque
from anytree import Node as Nd


class Node:
	def __init__(self, data):
		self.data = data
		self.parent = None
		self.children = [] 

	def add_child(self, obj):
		self.children.append(obj)
		obj.parent = self

	def preorder(self):
		root = self

		root_node = Nd(root.data['title'], **{'node_id' : root.data['id']})
		def dfs(node: Node, root_node: Nd, parent: Nd) -> None:
			if not node: return
			elif parent is None:
				parent = root_node
			else:
				nd = Nd(node.data['title'], parent, **{'node_id' : node.data['id']})
				parent = nd
			for c in node.children:
				dfs(c, root_node, parent)

		dfs(root, root_node, None)
		return root_node 
	
	def is_folder(node):
		key = 'contentHandler'
		from bbcli.utils.content_handler import content_handler
		return key in node.data and node.data[key]['id'] == content_handler['folder']

	# This is only getting folders. 
	def preorder2(self):
		root = self
		root_node = Nd(root.data['title'])	
		def dfs(node, root_node, parent):
			if not node: return
			elif parent is None:
				parent = root_node
			elif len(node.children) == 0 and Node.is_folder(node):
				# print("children0", node.data['title'])
				node.parent.children.remove(node)
			else:
				nd = Nd(node.data['title'], parent)
				parent = nd
			for c in node.children:
				dfs(c, root_node, parent)

		if len(root.children) == 0 and Node.is_folder(root):
			return None
		else:
			dfs(root, root_node, None)
			return root_node
		

	def __str__(self, level=0):
		ret = '\t'*level+self.data['title']+'\n'
		for child in self.children:
			ret += child.__str__(level+1)
		return ret



from dataclasses import dataclass
from datetime import date, datetime


"""
OPTIONS DATA CLASSES
"""


@dataclass
class DateInterval:
    start_date: date = None
    end_date: date = None


@dataclass
class StandardOptions:
    hide_content: bool = False
    reviewable: bool = False
    date_interval: DateInterval = DateInterval()


@dataclass
class FileOptions:
    launch_in_new_window: bool = False


@dataclass
class WeblinkOptions:
    launch_in_new_window: bool = False


@dataclass
class GradingOptions:
    due: datetime = None
    attempts_allowed: int = 0
    grade_schema_id: str = ''
    is_unlimited_attemps_allowed: bool = True
    score_possible: int = 0


"""
CONTENT-TYPE DATA CLASSES
"""


@dataclass
class FileContent:
    upload_id: str
    file_name: str
    mime_type: str
    # Options are Rename, Replace, ThrowError
    duplicate_file_handling: str = 'Rename'



import json
from typing import Dict, Any, List
import requests
from datetime import date

from bbcli.utils.URL_builder import URL_builder

url_builder = URL_builder()


def list_courses(session: requests.Session, user_name: str) -> Any:

    terms = get_terms(session)
    sort_terms(terms)

    term_1 = terms[len(terms) - 1]
    term_2 = terms[len(terms) - 2]

    course_memberships = get_course_memberships(session, user_name)

    course_list = []

    # Get courses from the correct terms
    for course in course_memberships:
        url = url_builder.base_v3().add_courses().add_id(
            course['courseId']).create()
        response = session.get(url, params={'fields': 'id, name, termId'})
        response.raise_for_status()
        response = json.loads(response.text)
        if response['termId'] == term_1['id'] or response['termId'] == term_2['id']:
            course_list.append({
                'id': response['id'],
                'name': response['name']
            })
        else:
            break

    return course_list


def list_all_courses(session: requests.Session, user_name: str) -> Any:
    course_memberships = get_course_memberships(session, user_name)

    course_list = []

    for course in course_memberships:
        url = url_builder.base_v3().add_courses().add_id(
            course['courseId']).create()
        response = session.get(url, params={'fields': 'id, name'})
        response.raise_for_status()
        response = json.loads(response.text)
        course_list.append(response)

    return course_list


def list_course(session: requests.Session, course_id: str) -> Any:
    url = url_builder.base_v3().add_courses().add_id(course_id).create()
    response = session.get(url)
    response.raise_for_status()

    return json.loads(response.text)


"""

HELPER FUNCTIONS

"""


def take_start_date(elem):
    return date.fromisoformat(elem['availability']['duration']['start'].split('T')[0])


def get_terms(session: requests.Session):
    url = url_builder.base_v1().add_terms().create()
    terms = session.get(url)
    terms.raise_for_status()
    terms = json.loads(terms.text)['results']
    return terms


def sort_terms(terms):
    # Sort terms by start date to get the two most recent semesters to determine which courses to show
    for term in terms:
        if term['availability']['duration']['type'] != 'DateRange':
            terms.remove(term)
    terms.sort(key=take_start_date)


def get_course_memberships(session: requests.Session, user_name: str):
    url = url_builder.base_v1().add_users().add_id(
        id=user_name, id_type='userName').add_courses().create()
    course_memberships = session.get(url)
    course_memberships.raise_for_status()
    course_memberships = json.loads(course_memberships.text)['results']
    return course_memberships



import click
from bbcli.services import courses_service
from bbcli.utils.error_handler import list_exception_handler
from bbcli.views import course_view
import os
import requests


#, help='List a spesific course with the corresponding id'
@click.command(name='list', help='This command lists your courses, by default only the courses from two last semesters.')
@click.option('-c', '--course', 'course_id', required=False, type=str, help='COURSE ID, search for spesific course')
@click.option('-a', '--all/--no-all', 'show_all', default=False, help='Lists all courses you have ever been signed up for')
@click.pass_context
@list_exception_handler
def list_courses(ctx, course_id=None, show_all=False):
    response = None

    if course_id:
        response = courses_service.list_course(
            session=ctx.obj['SESSION'], course_id=course_id)
        course_view.print_course(response)
    else:
        user_name = os.getenv('BB_USERNAME')
        if show_all:
            response = courses_service.list_all_courses(
                session=ctx.obj['SESSION'], user_name=user_name)
        else:
            response = courses_service.list_courses(
                session=ctx.obj['SESSION'], user_name=user_name)
        course_view.print_courses(response)



import click
import webbrowser

from bbcli.services import contents_service
from bbcli.utils.utils import check_response, html_to_text
from bbcli.entities.Node import Node
from bbcli.utils.content_handler import content_handler
from bbcli.views import contents_view

def is_folder(node):
    key = 'contentHandler'
    return key in node and node[key]['id'] == content_handler['folder']

def get_children(ctx, course_id, worklist, folder_ids, node_ids):
    if len(worklist) == 0:
        return
    else:
        node = worklist.pop(0)
        node_id = node.data['id']
        response = contents_service.get_children(
            ctx.obj['SESSION'], course_id, node_id)
        if check_response(response) == False:
            pass
        else:
            children = response.json()['results']
            for child in children:
                child_node = Node(child)
                node.add_child(child_node)
                if is_folder(child):
                    worklist.append(child_node)
                    folder_ids.append(child['id'])
                else:
                    node_ids.append(child['id'])

            return get_children(ctx, course_id, worklist, folder_ids, node_ids)


def get_folders(ctx, course_id, worklist, folder_ids, node_ids):
    if len(worklist) == 0:
        return
    else:
        node = worklist.pop(0)
        node_id = node.data['id']
        response = contents_service.get_children(
            ctx.obj['SESSION'], course_id, node_id)
        if check_response(response) == False:
            pass
        else:
            children = response.json()['results']
            for child in children:
                if is_folder(child):
                    child_node = Node(child)
                    node.add_child(child_node)
                    worklist.append(child_node)
                    folder_ids.append(child['id'])

            return get_folders(ctx, course_id, worklist, folder_ids, node_ids)


def get_content_type(ctx, course_id, worklist, folder_ids, node_ids, content_type):
    if len(worklist) == 0:
        return
    else:
        node = worklist.pop(0)
        node_id = node.data['id']
        response = contents_service.get_children(
            ctx.obj['SESSION'], course_id, node_id)
        if check_response(response) == False:
            pass
        else:
            children = response.json()['results']
            for child in children:
                if is_folder(child):
                    child_node = Node(child)
                    node.add_child(child_node)
                    worklist.append(child_node)
                    folder_ids.append(child['id'])
                elif 'contentHandler' in child and content_handler[content_type] == child['contentHandler']['id']:
                    child_node = Node(child)
                    node.add_child(child_node)
                    node_ids.append(child['id'])

            return get_content_type(
                ctx, course_id, worklist, folder_ids, node_ids, content_type)


def list_contents_thread(
        ctx,
        course_id,
        worklist,
        folder_ids,
        node_ids,
        root,
        folders: bool,
        content_type: str):
    if folders and content_type is not None:
        click.ClickException(
            'Cannot list contents and a specific folder type. Try either one.')
    elif folders and content_type is None:
        get_folders(ctx, course_id, worklist, folder_ids, node_ids)
    elif folders == False and content_type is not None:
        get_content_type(ctx, course_id, worklist, folder_ids, node_ids,  content_type)
    else:
        get_children(ctx, course_id, worklist, folder_ids, node_ids)

    root_node = root.preorder()
    return root_node


def check_content_handler(ctx, course_id: str, node_id: str):
    session = ctx.obj['SESSION']
    response = contents_service.get_content(
        ctx.obj['SESSION'], course_id, node_id)
    if check_response(response) == False:
        return
    data = response.json()
    ch = data['contentHandler']['id']
    if ch == content_handler['document']:
        response = contents_service.get_attachments(session, course_id, node_id)
        attachments = response.json()['results']
        str = data['title'] + '\n'
        body = '' if 'body' not in data else html_to_text(data['body'])
        str += body
        contents_view.open_less_page(str)
        if len(attachments) > 0:
            click.confirm(
                "This is a document with an attachment(s), do you want to download it?",
                abort=True)
            paths = contents_service.download_attachments(session, course_id, node_id, attachments)
            [contents_service.open_file(path) for path in paths]
        else:
            click.echo('The document has no attachments.')
    elif ch == content_handler['externallink']:
        link = data['contentHandler']['url']
        click.echo(f'Opening an weblink: {link}')
        webbrowser.open(link)
    elif ch == content_handler['folder']: # this will list the contents of a folder
        click.echo('Listing the contents of a folder...')
        folder_ids = []
        node_ids = []
        folder_ids.append(data['id'])
        root = Node(data)
        worklist = [root]
        get_children(ctx, course_id, worklist, folder_ids, node_ids)
        root_node = root.preorder()
        contents_view.list_tree(root_node, folder_ids, node_ids)
    elif ch == content_handler['courselink']:
        click.echo('Opening the contents of a courselink...')
        key = 'targetId'
        if key in data['contentHandler']:
            target_id = data['contentHandler'][key]
            check_content_handler(ctx, course_id, target_id)
    elif ch == content_handler['file']:
        response = contents_service.get_attachments(
            session, course_id, node_id)
        attachments = response.json()['results']
        if len(attachments) > 0:
            click.confirm(
                "This is a file, do you want to download and open it?",
                abort=True)
        paths = contents_service.download_attachments(
            session, course_id, node_id, attachments)
        [contents_service.open_file(path) for path in paths]
    elif ch == content_handler['assignment']:
        click.echo('Opening assignment...')
        str = data['title'] + '\n' + html_to_text(data['body'])
        contents_view.open_less_page(str)
        response = contents_service.get_attachments(
            session, course_id, node_id)
        attachments = response.json()['results']
        if len(attachments) > 0:
            click.confirm(
                "The assignment contains attachment(s), do you want to download?",
                abort=True)
            paths = contents_service.download_attachments(
                session, course_id, node_id, attachments)
            [contents_service.open_file(path) for path in paths]
    elif ch == content_handler['blankpage']:
        click.echo('Opening blankpage...')
        str = data['title'] + '\n' + html_to_text(data['body'])
        contents_view.open_less_page(str)
    else:
        click.echo('The cli does not currently support the content type.')


image_files = ['jpeg', 'jpg', 'gif', 'svg', 'png', 'tiff', 'tif']
document_files = ['pdf', 'doc', 'docx', 'html', 'htm', 'xls', 'xlsx', 'txt']
video_files = ['mp4', 'avi', 'mov', 'flv', 'avchd']
presentation_files = ['ppt', 'pptx', 'odp', 'key']
audio_files = ['m4a', 'mp3', 'wav']


def get_file_type(fn):
    ft = ''
    ft = ['image_file' for imf in image_files if imf in fn]
    if len(ft) > 0:
        return ft[0]
    ft = ['document_file' for df in document_files if df in fn]
    if len(ft) > 0:
        return ft[0]
    ft = ['video_files' for vf in video_files if vf in fn]
    if len(ft) > 0:
        return ft[0]
    ft = ['presentation_files' for pf in presentation_files if pf in fn]
    if len(ft) > 0:
        return ft[0]
    ft = ['audio_files' for af in audio_files if af in fn]
    if len(ft) > 0:
        return ft[0]
    else:
        print("Could not find any of the files")
        # TODO: maybe throw something
        return



import json
import os
import requests
import magic
from bbcli.utils.URL_builder import Builder, URL_builder
from bbcli.services.utils.content_builder import ContentBuilder
from bbcli.entities.content_builder_entitites import FileContent, GradingOptions, StandardOptions, FileOptions, WeblinkOptions
from bbcli.utils.utils import input_body
import click
import webbrowser

from bbcli.utils.utils import check_response, get_download_path

url_builder = URL_builder()
content_builder = ContentBuilder()

# User gets a tree structure view of the courses content
# where each content is listed something like this: _030303_1 Lectures Folder
def list_contents(session: requests.Session, course_id):
    url = url_builder.base_v1().add_courses().add_id(course_id).add_contents().create()
    response = session.get(url)
    response.raise_for_status()
    return response

# get the children of a specific folder
def get_children(session: requests.Session, course_id: str, node_id: str):
    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_contents().add_id(node_id).add_children().create()
    response = session.get(url)
    response.raise_for_status()
    return response

# If it is a folder, list it like a tree structure view like mentioned above.
# If it is a document, download and open the document maybe?
# Find all types of content and have an appropriate response for them. This
# should maybe be handled in the view...


def get_content(session: requests.Session, course_id: str, node_id: str):
    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_contents().add_id(node_id).create()
    response = session.get(url)
    response.raise_for_status()
    return response

def get_content_targetid(session: requests.Session, course_id: str, target_id: str):
    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_contents().add_id(target_id).create()
    return session.get(url)

def get_attachments(session: requests.Session, course_id: str, node_id: str):
    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_contents().add_id(node_id).add_attachments().create()
    response = session.get(url)
    response.raise_for_status()
    return response

def download_attachment(session: requests.Session, course_id: str, node_id: str, attachment) -> str:
    attachment_id = attachment['id']
    fn = attachment['fileName']
    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_contents().add_id(node_id).add_attachments(
        ).add_id(attachment_id).add_download().create()
    response = session.get(url, allow_redirects=True)
    response.raise_for_status()
    downloads_path = get_download_path(fn)
    f = open(downloads_path, 'wb')
    f.write(response.content)
    f.close()
    click.echo(f'\"{fn}\" was downloaded to the downloads folder.')
    return downloads_path 

def download_attachments(session: requests.Session, course_id: str, node_id: str, attachments):
    paths = []
    for attachment in attachments:
        path = download_attachment(session, course_id, node_id, attachment)
        paths.append(path)
    return paths

def open_file(path):
    webbrowser.open(r'file:'+path)

def upload_attachment(session: requests.Session, course_id: str, content_id: str, file_dst: str):
    uploaded_file = upload_file(session, file_dst)
    data = json.dumps(uploaded_file)

    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_contents().add_id(content_id).add_attachments().create()
    response = session.post(url, data=data)
    response.raise_for_status()


def create_document(session: requests.Session, course_id: str, parent_id: str, title: str, standard_options: StandardOptions = None, attachments: tuple = None):

    data_body = input_body()
    data = content_builder\
        .add_parent_id(parent_id)\
        .add_title(title)\
        .add_body(data_body)\
        .add_standard_options(standard_options)\
        .add_content_handler_document()\
        .create()

    data = json.dumps(data)
    url = generate_create_content_url(course_id, parent_id)
    response = session.post(url, data=data)
    response.raise_for_status()

    created_content_id = json.loads(response.text)['id']
    handle_attachments(session, course_id, created_content_id, attachments)

    return response.text

# TODO: Bug that if a file is created with an attachment, the attachment takes the place of the actual file for the content. In addition,
#       if two attachments is added, only the last one is added/overwrite the first one


def create_file(session: requests.Session, course_id: str, parent_id: str, title: str, file_dst: str, file_options: FileOptions, standard_options: StandardOptions):

    uploaded_file = upload_file(session, file_dst)
    mime = magic.Magic(mime=True)
    mime_type = mime.from_file(file_dst)

    with open(file_dst, 'rb') as f:
        file_name = os.path.basename(f.name)

    file_content = FileContent(uploaded_file['id'], file_name, mime_type)

    data = content_builder\
        .add_parent_id(parent_id)\
        .add_title(title)\
        .add_standard_options(standard_options)\
        .add_file_options(file_options)\
        .add_content_handler_file(file_content)\
        .create()

    data = json.dumps(data)
    url = generate_create_content_url(course_id, parent_id)
    response = session.post(url, data=data)
    response.raise_for_status()

    return response.text


def create_externallink(session: requests.Session, course_id: str, parent_id: str, title: str, url: str, web_link_options: WeblinkOptions, standard_options: StandardOptions):

    data = content_builder\
        .add_parent_id(parent_id)\
        .add_title(title)\
        .add_standard_options(standard_options)\
        .add_weblink_options(web_link_options)\
        .add_content_handler_externallink(url)\
        .create()

    data = json.dumps(data)
    url = generate_create_content_url(course_id, parent_id)
    response = session.post(url, data=data)
    response.raise_for_status()
    return response.text


def create_folder(session: requests.Session, course_id: str, parent_id: str, title: str, is_bb_page: bool, standard_options: StandardOptions):

    data_body = input_body()

    data = content_builder\
        .add_title(title)\
        .add_body(data_body)\
        .add_standard_options(standard_options)\
        .add_content_handler_folder(is_bb_page=is_bb_page)

    if parent_id:
        url = generate_create_content_url(course_id, parent_id)
        data.add_parent_id(parent_id)
    else:
        url = url_builder.base_v1().add_courses().add_id(
            course_id).add_contents().create()
    data = data.create()

    data = json.dumps(data)
    response = session.post(url, data=data)
    response.raise_for_status()
    return response.text


# TODO:FUNKER IKKE PGA targetType
def create_courselink(session: requests.Session, course_id: str, parent_id: str, title: str, target_id: str, standard_options: StandardOptions):

    data_body = input_body()

    data = content_builder\
        .add_title(title)\
        .add_body(data_body)\
        .add_standard_options(standard_options)\
        .add_content_handler_courselink(target_id=target_id)\
        .create()

    data = json.dumps(data)
    url = generate_create_content_url(course_id, parent_id)

    response = session.post(url, data=data)
    response.raise_for_status()
    return response.text

# TODO: Figure out how a lecturer can get/edit submission-, grading-, and display of grades options.


def create_assignment(session: requests.Session, course_id: str, parent_id: str, title: str, standard_options: StandardOptions, grading_options: GradingOptions, attachments: tuple = None):

    instructions = input_body()

    data = content_builder\
        .add_parent_id(parent_id)\
        .add_title(title)\
        .add_standard_options(standard_options)\
        .add_grading_options(grading_options)\
        .create()

    data.update({'instructions': instructions})

    if attachments:
        files = []
        for attachment in attachments:
            files.append(upload_file(session, attachment)['id'])
        data.update({'fileUploadIds': files})

    data = json.dumps(data)

    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_contents().add_create_assignment().create()
    response = session.post(url, data=data)
    response.raise_for_status()
    return response.text


def delete_content(session: requests.Session, course_id: str, content_id: str, delete_grades: bool):
    parameters = {
        'deleteGrades': delete_grades
    }
    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_contents().add_id(content_id).create()
    response = session.delete(url, params=parameters)
    response.raise_for_status()
    return response.text


def update_content(session: requests.Session, course_id: str, content_id: str):
    url = url_builder.base_v1().add_courses().add_id(
        course_id).add_contents().add_id(content_id).create()
    content = session.get(url)
    content = json.loads(content.text)
    if not is_editable_content_type(content['contentHandler']['id']):
        click.echo('This content type is not editable')
        raise click.Abort()
    if 'contentHandler' in content:
        del content['contentHandler']
    if 'links' in content:
        del content['links']
    MARKER = '# Everything below is ignored.\n'
    editable_data = json.dumps(content, indent=2)
    new_data = click.edit(editable_data + '\n\n' + MARKER)

    response = session.patch(url, data=new_data)
    response.raise_for_status()
    return response.text


"""

HELPER FUNCTIONS

"""


def upload_file(session: requests.Session, dst: str):

    del session.headers['Content-Type']
    with open(dst, 'rb') as f:
        file_name = os.path.basename(f.name)
        files = {file_name: f.read()}

    url = url_builder.base_v1().add_uploads().create()
    response = session.post(url, files=files)
    response.raise_for_status()
    session.headers.update({
        'Content-Type': 'application/json'
    })
    file = json.loads(response.text)
    return file


def generate_create_content_url(course_id: str, content_id: str):
    return url_builder\
        .base_v1()\
        .add_courses()\
        .add_id(course_id)\
        .add_contents()\
        .add_id(content_id)\
        .add_children()\
        .create()


def handle_attachments(session: requests.Session, course_id: str, content_id: str, attachments: tuple or None):
    if attachments:
        for attachment in attachments:
            upload_attachment(session, course_id, content_id, attachment)


def is_editable_content_type(content_type: str):
    valid_content_types = ['resource/x-bb-assignment', 'resource/x-bb-externallink',
                           'resource/x-bb-courselink', 'resource/x-bb-file', 'resource/x-bb-document']
    for type in valid_content_types:
        if content_type == type:
            return True
    return False



import os
from setuptools import find_packages

here = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(here, 'bbcli', '__version__.py'), 'r') as f:
    exec(f.read(), about)

with open('README.md', 'r') as f:
    readme = f.read()

requires = [
    'Click>=8.1',
    'colorama',
    'requests',
    'python-dotenv',
    'beautifulsoup4',
    'lxml>=4.8.0',
    'shellingham',
    'anytree',
    'html2text',
    'python-magic',
    'python-dateutil',
    'tabulate==0.8.9'
]

def setup_package():
    metadata = dict(
        name=about['__title__'],
        version=about['__version__'],
        description=about['__description__'],
        long_description=readme,
        long_description_content_type='text/markdown',
        url=about['__url__'],
        author=about['__authors__'],
        author_email=about['__author_emails__'],
        license=about['__license__'],
        packages=find_packages(),
        include_package_data=True,
        install_requires=requires,
        entry_points={
            'console_scripts': [
                'bb=bbcli.__main__:main', # the cli() function runs inside the bbcli.py. cli is the command that is used to run the command
            ],
        }
    )

    try:
        from setuptools import setup
    except ImportError:
        from distutils.core import setup
    
    setup(**metadata)

if __name__ == '__main__':
    setup_package()



"""Top-level package for RP To-Do."""
# rptodo/__init__.py

__app_name__ = "bbcli"
__version__ = "0.1.0"
# from .endpoints import *
# from .entities.Node import *
from .utils.utils import *
from .services import login
# from endpoints import get_user, get_course, get_assignments, get_course_contents

(
    SUCCESS,
    DIR_ERROR,
    FILE_ERROR,
    DB_READ_ERROR,
    DB_WRITE_ERROR,
    JSON_ERROR,
    ID_ERROR,
) = range(7)

ERRORS = {
    DIR_ERROR: "config directory error",
    FILE_ERROR: "config file error",
    DB_READ_ERROR: "database read error",
    DB_WRITE_ERROR: "database write error",
    ID_ERROR: "to-do id error",
}



VERSION = (0, 0, 9)
PRERELEASE = None  # alpha, beta or rc
REVISION = None


def generate_version(version, prerelease=None, revision=None):
    version_parts = ['.'.join(map(str, version))]
    if prerelease is not None:
        version_parts.append('-{}'.format(prerelease))
    if revision is not None:
        version_parts.append('.{}'.format(revision))
    return ''.join(version_parts)

__title__ = 'Blackboard LMS CLI'
__description__ = 'A command-line tool suite for communicating with the Blackboard Learn Managment System'
__url__ = 'https://gitlab.stud.idi.ntnu.no/mattiaae/idatt2900-072'
__version__ = generate_version(VERSION, prerelease=PRERELEASE, revision=REVISION)
__authors__ = 'Hans William Forebrigd, Mattias Aggentoft Eggen, Magnus Bredeli'
__author_emails__ = 'hansw0701@gmail.com, mattias.a.eggen@gmail.com, magnus.bredeli@hotmail.com'
__license__ = 'Willyeh Wonkmeh 2.0'


from __future__ import annotations
from datetime import date, datetime
from typing import Any, Dict
from abc import ABC, abstractmethod

from bbcli.entities.content_builder_entitites import FileContent, FileOptions, GradingOptions, StandardOptions, WeblinkOptions


class Builder(ABC):

    @property
    @abstractmethod
    def product(self) -> None:
        pass

    @abstractmethod
    def add_parent_id(self, title: str) -> Builder:
        pass

    @abstractmethod
    def add_title(self, title: str) -> Builder:
        pass

    @abstractmethod
    def add_body(self, body: str) -> Builder:
        pass

    @abstractmethod
    def add_standard_options(self, standard_options: StandardOptions) -> Builder:
        pass

    # Alignment option is available in creation in the web interface, but not in the actual content objects that is created
    @abstractmethod
    def add_file_options(self, file_options: FileOptions) -> Builder:
        pass

    @abstractmethod
    def add_weblink_options(self, web_link_options: WeblinkOptions) -> Builder:
        pass

    @abstractmethod
    def add_grading_options(self, grading_options: GradingOptions) -> Builder:
        pass

    @abstractmethod
    def add_content_handler_document(self) -> Builder:
        pass

    @abstractmethod
    def add_content_handler_file(self, file_content: FileContent) -> Builder:
        pass

    @abstractmethod
    def add_content_handler_externallink(self, url: str) -> Builder:
        pass

    @abstractmethod
    def add_content_handler_folder(self, is_bb_page: bool) -> Builder:
        pass

    # Possible target types:
    # Unset
    # CourseAssessment
    # CourseTOC
    # Forum
    # Tool
    # CollabSession (deprecated since 3000.1.0)
    # Group
    # BlogJournal
    # StaffInfo
    # ModulePage

    @abstractmethod
    def add_content_handler_courselink(self, target_id: str, target_type: str = 'Unset') -> Builder:
        pass


class ContentBuilder(Builder):

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        self._product = Content()

    @property
    def product(self) -> Content:
        product = self._product
        self.reset()
        return product

    def add_parent_id(self, parent_id: str) -> Builder:
        self._product.add({
            'parentId': parent_id
        })
        return self

    def add_title(self, title: str) -> Builder:
        self._product.add({
            'title': title
        })
        return self

    def add_name(self, name: str) -> Builder:
        self._product.add({
            'name': name
        })
        return self

    def add_body(self, body: str) -> Builder:
        self._product.add({
            'body': body
        })
        return self

    def add_standard_options(self, standard_options: StandardOptions) -> Builder:
        start_date_str = datetime.strftime(standard_options.date_interval.start_date,
                                           '%Y-%m-%dT%H:%m:%S.%fZ') if standard_options.date_interval.start_date else None
        end_date_str = datetime.strftime(standard_options.date_interval.end_date,
                                         '%Y-%m-%dT%H:%m:%S.%fZ') if standard_options.date_interval.end_date else None
        self._product.add({
            'availability': {
                'available': 'No' if standard_options.hide_content else 'Yes',
                'allowGuests': True,
                'allowObservers': True,
                'adaptiveRelease': {
                    'start': start_date_str,
                    'end': end_date_str
                }
            }
        })
        if standard_options.reviewable:
            self._product.add({
                'reviewable': standard_options.reviewable,
            })
        return self

    # Missing an extra option, but don't know what it is called
    def add_file_options(self, file_options: FileOptions) -> Builder:
        self._product.add({
            'launchInNewWindow': file_options.launch_in_new_window
        })
        return self

    def add_weblink_options(self, web_link_options: WeblinkOptions) -> Builder:
        self._product.add({
            'launchInNewWindow': web_link_options.launch_in_new_window
        })
        return self

    def add_grading_options(self, grading_options: GradingOptions) -> Builder:
        due_date_str = datetime.strftime(
            grading_options.due, '%Y-%m-%dT%H:%m:%S.%fZ') if grading_options.due else None
        self._product.add({
            'grading': {
                'due': due_date_str,
                'attemptsAllowed': grading_options.attempts_allowed,
                'isUnlimitedAttemptsAllowed': grading_options.is_unlimited_attemps_allowed
            },
            'score': {
                'possible': grading_options.score_possible
            }
        })
        return self

    def add_content_handler_document(self) -> Builder:
        self._product.add({
            'contentHandler': {
                'id': 'resource/x-bb-document'
            }
        })
        return self

    def add_content_handler_file(self, file_content: FileContent) -> Builder:
        self._product.add({
            'contentHandler': {
                'id': 'resource/x-bb-file',
                'file': {
                    'uploadId': file_content.upload_id,
                    'fileName': file_content.file_name,
                    'mimeType': file_content.mime_type,
                    'duplicateFileHandling': file_content.duplicate_file_handling
                }

            }
        })
        return self

    def add_content_handler_externallink(self, url: str) -> Builder:
        self._product.add({
            'contentHandler': {
                'id': 'resource/x-bb-externallink',
                'url': url
            }
        })
        return self

    def add_content_handler_folder(self, is_bb_page: bool) -> Builder:
        self._product.add({
            'contentHandler': {
                'id': 'resource/x-bb-folder',
                'isBbPage': is_bb_page
            }
        })
        return self

    def add_content_handler_courselink(self, target_id: str, target_type: str = 'Unset') -> Builder:
        self._product.add({
            'contentHandler': {
                'id': 'resource/x-bb-courselink',
                'targetId': target_id,
                'targetType': target_type
            }
        })
        return self

    def create(self) -> Dict:
        content = self._product.get_content()
        self._product = Content()
        return content


class Content():

    def __init__(self) -> None:
        self.content = {}

    def add(self, content_part: str) -> None:
        self.content.update(content_part)

    def get_content(self) -> None:
        return self.content



from __future__ import annotations
from typing import Any
from abc import ABC, abstractmethod

DOMAIN = 'https://ntnu.blackboard.com'
API_BASE = '/learn/api/public'


class Builder(ABC):

    @property
    @abstractmethod
    def product(self) -> None:
        pass

    """
    Returns the base URL which includes the domain and first part of all the endpoints: domain/learn/api/public/vX,
    where X is the version from 1 to 3.
    """

    @abstractmethod
    def base_v1(self) -> Builder:
        pass

    @abstractmethod
    def base_v2(self) -> Builder:
        pass

    @abstractmethod
    def base_v3(self) -> Builder:
        pass

    @abstractmethod
    def add_courses(self) -> Builder:
        pass

    @abstractmethod
    def add_users(self) -> Builder:
        pass

    @abstractmethod
    def add_announcements(self) -> Builder:
        pass

    @abstractmethod
    def add_contents(self) -> Builder:
        pass

    @abstractmethod
    def add_terms(self) -> Builder:
        pass

    @abstractmethod
    def add_children(self) -> Builder:
        pass

    @abstractmethod
    def add_attachments(self) -> Builder:
        pass

    @abstractmethod
    def add_uploads(self) -> Builder:
        pass

    @abstractmethod
    def add_create_assignment(self) -> Builder:
        pass

    @abstractmethod
    def add_id(self, id: str, id_type: str = None) -> Builder:
        pass

    @abstractmethod
    def add_download(self) -> Builder:
        pass

    @abstractmethod
    def add_gradebook(self) -> Builder:
        pass

    @abstractmethod
    def add_columns(self) -> Builder:
        pass

    @abstractmethod
    def add_attempts(self) -> Builder:
        pass

    @abstractmethod
    def add_files(self) -> Builder:
        pass

    @abstractmethod
    def create(self) -> Builder:
        pass


class URL_builder(Builder):

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        self._product = URL()

    @property
    def product(self) -> URL:
        product = self._product
        self.reset()
        return product

    def base_v1(self) -> URL_builder:
        self._product.add(f'{DOMAIN}{API_BASE}/v1')
        return self

    def base_v2(self) -> URL_builder:
        self._product.add(f'{DOMAIN}{API_BASE}/v2')
        return self

    def base_v3(self) -> URL_builder:
        self._product.add(f'{DOMAIN}{API_BASE}/v3')
        return self

    def add_courses(self) -> URL_builder:
        self._product.add('/courses')
        return self

    def add_users(self) -> URL_builder:
        self._product.add('/users')
        return self

    def add_announcements(self) -> URL_builder:
        self._product.add('/announcements')
        return self

    def add_contents(self) -> URL_builder:
        self._product.add('/contents')
        return self

    def add_terms(self) -> URL_builder:
        self._product.add('/terms')
        return self

    def add_children(self) -> Builder:
        self._product.add('/children')
        return self

    def add_attachments(self) -> Builder:
        self._product.add('/attachments')
        return self

    def add_uploads(self) -> Builder:
        self._product.add('/uploads')
        return self

    def add_create_assignment(self) -> Builder:
        self._product.add('/createAssignment')
        return self

    def add_download(self) -> Builder:
        self._product.add('/download')
        return self

    def add_id(self, id: str, id_type: str = None) -> URL_builder:
        if id_type:
            self._product.add(f'/{id_type}:{id}')
        else:
            self._product.add(f'/{id}')
        return self

    def add_gradebook(self) -> Builder:
        self._product.add('/gradebook')
        return self

    def add_columns(self) -> Builder:
        self._product.add('/columns')
        return self

    def add_attempts(self) -> Builder:
        self._product.add('/attempts')
        return self
    
    def add_files(self) -> Builder:
        self._product.add('/files')
        return self

    def create(self) -> str:
        url = self._product.get_url()
        self._product = URL()
        return url


class URL():

    def __init__(self) -> None:
        self.URL = ''

    def add(self, url_part: str) -> None:
        self.URL += url_part

    def get_url(self) -> None:
        return self.URL
